#  WotoPlatform-Py - A Python library for interacting with WotoPlatform API.
#  Copyright (C) 2021-2022  ALiwoto - <woto@kaizoku.cyou> <https://github.com/ALiwoto>
#
#  This file is part of WotoPlatform-Py.
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import json
import inspect
import logging
from typing import Callable, Union
import uuid

from wotoplatform.types.woto_crypto.password_container import PasswordContainer256
from .utils import (
    WotoSocket,
)
from .tools import (
    DataReceiver,
    make_sure_byte, 
)
from .types.errors import (
    WrongUsername,
    InvalidTypeException,
    ClientVersionNotAcceptable,
    ClientNotInitializedException,
    ClientAlreadyInitializedException,
)
from .types import (
    RawResponse,
    RawDScaffold,
    ScaffoldHolder,
    ClientBase,
    DScaffold,
    RScaffold,
    Scaffold,
    VersionData,
    VersionResponse,
    RegisterUserData,
    LoginUserData, 
    LoginUserResult,
    RegisterUserResponse, 
    RegisterUserResult,
    GetUserFavoriteData,
    GetUserFavoriteResult,
    GetUserFavoriteCountResult,
    GetUserFavoriteCountData,
    ChangeNamesData, 
    ChangeUserBioData, 
    GetMeData, 
    GetMeResult, 
    GetUserInfoData, 
    GetUserInfoResult, 
    SetUserFavoriteData,
    DeleteUserFavoriteData,
    ResolveUsernameData,
)

__version__ = '0.0.21'
log = logging.getLogger(__name__)

class WotoClient(ClientBase):
    username: str = ''
    __password: str = ''
    endpoint_url: str = ''
    auth_key: str = ''
    access_hash: str = ''
    is_initialized: bool = False
    is_logged_in: bool = False
    client_version: VersionData = None
    connection_closed_handler: Callable = None
    use_tls: bool = True
    __endpoint: str = ''
    __port: int = 0
    __woto_socket: WotoSocket = None
    __MAX_DATA_BUFFER = 8
    __internal_receiver = {}
    __read_task: asyncio.Task = None
    #__internal_loop = None
    __read_data_error: Exception = None
    

    def __init__(
        self, 
        username: str, 
        password: str, 
        endpoint: str = 'wotoplatform.kaizoku.cyou', 
        port: int = 50100,
        use_tls: bool = True
    ):
        if not username:
            raise ValueError('username cannot be empty')
        if not password:
            raise ValueError('password cannot be empty')
        
        self.username = username
        self.__password = password
        self.__endpoint = endpoint
        self.__port = port

        self.__woto_socket = WotoSocket(
            host=endpoint, 
            port=port,
            use_tls=use_tls,
            ipv6=False
        )
    
    def __del__(self) -> None:
        if self.__read_task and not self.__read_task.done():
            try:
                self.__read_task.cancel()
            except Exception: pass
        
    async def start(self) -> None:
        if self.is_initialized:
            raise ClientAlreadyInitializedException()
        
        if not self.__woto_socket:
            self.__woto_socket = WotoSocket(host=self.__endpoint, port=self.__port)
        
        if self.__read_task and not self.__read_task.done():
            try:
                self.__read_task.cancel()
            except Exception: pass
        
        self.__read_task = asyncio.create_task(self.__read_data_loop())
        
        self.client_version = VersionData()
        self.is_initialized = True
        try:
            version_response = await self.send_and_parse(self.client_version)
            if not isinstance(version_response, VersionResponse):
                raise InvalidTypeException(VersionResponse, type(version_response))
            
            if not version_response.success:
                raise version_response.get_exception()
            
            if not version_response.result.is_acceptable:
                raise ClientVersionNotAcceptable()
            
            try:
                await self._login(
                    username=self.username,
                    password=self.__password,
                    auth_key=self.auth_key,
                    access_hash=self.access_hash,
                )
            except WrongUsername:
                await self._register(
                    username=self.username,
                    password=self.__password,
                )
            
            self.is_logged_in = True
        except:
            self.is_initialized = False
            raise
        
        # for security reasons, set the password to None here.
        self.__password = None
    
    async def stop(self) -> None:
        self.is_initialized = False
        self.is_logged_in = False
        if self.__woto_socket:
            await self.__woto_socket.close()
            self.__woto_socket = None
        
        if self.__read_task and not self.__read_task.done():
            try: self.__read_task.cancel()
            except Exception: pass
            try: await self.__read_task
            except asyncio.CancelledError: pass
        
        
    
    async def _login(self, username: str, password: str, auth_key:str, access_hash: str) -> LoginUserResult:
        """
        Login user. Don't use this method directly, instead use start method.
        """
        
        pass_container = PasswordContainer256()
        pass_container.set_as_password(password)
        response = await self.send_and_parse(
            LoginUserData(
                username=username,
                password=pass_container,
                auth_key=auth_key,
                access_hash=access_hash,
            )
        )

        if not response.success:
            raise response.get_exception()
        
        return response.result

    async def _register(self, username: str, password: str) -> RegisterUserResult:
        """
        Registers a new user on woto-platform. 
        Don't use this method directly, instead use start method.
        """
        pass_container = PasswordContainer256()
        pass_container.set_as_password(password)
        response = await self.send_and_parse(
            RegisterUserData(
                username=username,
                password=pass_container,
            )
        )
        
        if not isinstance(response, RegisterUserResponse):
            raise InvalidTypeException(RegisterUserResponse, type(response))

        if not response.success:
            raise response.get_exception()
        

        return response.result
    
    def get_read_data_error(self) -> Exception:
        return self.__read_data_error
    
    async def __read_data_loop(self) -> None:
        #await self.__woto_socket.connect_OLD(self.__internal_loop)
        await self.__woto_socket.connect()
        data = None
        while self.is_initialized:
            try:
                data = await self._read_data()
            except Exception as e:
                log.error(e, exc_info=True)
                self.__read_data_error = e
                data = None
                continue
            
            if not data:
                continue
            j_value = json.loads(data)
            j_uid = str(j_value['unique_id'])
            data_receiver = self.__internal_receiver.get(j_uid, None)
            if not isinstance(data_receiver, DataReceiver):
                #TODO: call handlers....
                continue
            
            data_receiver.receive_data(j_value)
            self.__internal_receiver.pop(j_uid, None)
            

    async def _write_data(self, data: bytes):
        bb = str(len(data)).zfill(self.__MAX_DATA_BUFFER)
        bb = make_sure_byte(bb, self.__MAX_DATA_BUFFER)
        await self.__woto_socket.send(bb + data)
    
    async def _read_data(self) -> bytes:
        count = await self.__woto_socket.recv(self.__MAX_DATA_BUFFER)
        if not count:
            if not self.is_initialized or not self.connection_closed_handler:
                return #TODO: logging
            self.connection_closed_handler()
            return
        count = int(count.decode('utf-8').strip())
        return await self.__woto_socket.recv(count)
    
    async def send(self, scaffold: Scaffold, timeout: float = 3):
        if not self.is_initialized:
            raise ClientNotInitializedException()
        
        if not isinstance(scaffold, Scaffold):
            raise InvalidTypeException(Scaffold, type(scaffold))

        uid = str(uuid.uuid4())
        holder = ScaffoldHolder(uid, scaffold)
        d_receiver = DataReceiver(self.__woto_socket)
        self.__internal_receiver[uid] = d_receiver
        # await self.client_lock.acquire()
        await d_receiver.first_wait()
        await self._write_data(holder.get_as_bytes())
        await d_receiver.wait_for_data(timeout)
        r_value = d_receiver.received_data
        # self.client_lock.release()
        return r_value
    
    async def send_raw_batch(self, action: int, batch_name: str, data, timeout: float = 3) -> RawResponse:
        if not isinstance(data, str):
            data = json.dumps(data)
        
        raw_scaffold = RawDScaffold(action, batch_name, data)
        return await self.send_and_parse(raw_scaffold, timeout=timeout)

    async def send_and_parse(self, scaffold: DScaffold, timeout: float = 10) -> RScaffold:
        if not isinstance(scaffold, DScaffold) and not isinstance(scaffold, RawDScaffold):
            return None
        
        response_type = scaffold.get_response_type()
        parser_method = getattr(scaffold, 'parse_response', None)
        if not response_type and not parser_method:
            return None
        
        j_value = await self.send(scaffold, timeout=timeout)
        if parser_method and inspect.ismethod(parser_method):
            return parser_method(j_value)
        
        res = response_type(**j_value)
        return res

    async def get_me(self) -> GetMeResult:
        response = await self.send_and_parse(GetMeData())

        if not response.success:
            raise response.get_exception()
        
        return response.result
    
    async def change_user_bio(self, bio: str, user_id: int = 0) -> bool:
        response = await self.send_and_parse(
            ChangeUserBioData(
                user_id=user_id,
                bio=bio,
            ),
        )

        if not response.success:
            raise response.get_exception()
        
        return response.result

    async def change_names(self, first_name: str, last_name: str, user_id: int = 0) -> bool:
        if not first_name and not last_name:
            return False
        
        response = await self.send_and_parse(
            ChangeNamesData(
                user_id=user_id,
                first_name=first_name,
                last_name=last_name,
            ),
        )

        if not response.success:
            raise response.get_exception()
        
        return response.result

    async def get_user_info(self, user_id: Union[int, str]) -> GetUserInfoResult:
        if isinstance(user_id, str):
            try:
                user_id = int(user_id)
            except ValueError: pass
        
        data: GetUserInfoData = None
        if isinstance(user_id, int):
            data = GetUserInfoData(user_id=user_id)
        elif isinstance(user_id, str):
            data = GetUserInfoData(username=user_id)
        else:
            raise InvalidTypeException(int, type(user_id))
        
        response = await self.send_and_parse(data)

        if not response.success:
            raise response.get_exception()
        
        return response.result

    async def get_user_favorite(self, key: str, user_id: int = 0) -> GetUserFavoriteResult:
        response = await self.send_and_parse(
            GetUserFavoriteData(
                user_id=user_id,
                favorite_key=key,
            ),
        )

        if not response.success:
            raise response.get_exception()
        
        return response.result
    
    async def get_user_favorite_value(self, key: str, user_id: int = 0) -> str:
        fav = await self.get_user_favorite(key, user_id)
        if isinstance(fav, GetUserFavoriteResult):
            return fav.favorite_value
        
        raise InvalidTypeException(GetUserFavoriteResult, type(fav))

    async def get_user_favorites_count(self, user_id: int = 0) -> int:
        response = await self.send_and_parse(
            GetUserFavoriteCountData(
                user_id=user_id,
            ),
        )

        if not response.success:
            raise response.get_exception()
        
        if isinstance(response.result, GetUserFavoriteCountResult):
            return response.result.favorites_count
        
        return 0

    async def set_user_favorite(self, key: str, value: str, user_id: int = 0) -> bool:
        response = await self.send_and_parse(
            SetUserFavoriteData(
                user_id=user_id,
                favorite_key=key,
                favorite_value=value,
            ),
        )

        if not response.success:
            raise response.get_exception()
        
        return response.result

    async def delete_user_favorite(self, key: str, user_id: int = 0) -> bool:
        if not isinstance(user_id, int):
            raise InvalidTypeException(int, type(user_id))
        
        response = await self.send_and_parse(
            DeleteUserFavoriteData(
                user_id=user_id,
                favorite_key=key,
            ),
        )

        if not response.success:
            raise response.get_exception()
        
        return response.result
    
    async def resolve_username(self, username: str) -> GetUserInfoResult:
        if not isinstance(username, str):
            raise InvalidTypeException(str, type(username))
        

        response = await self.send_and_parse(
            ResolveUsernameData(
                username=username,
            ),
        )

        if not response.success:
            raise response.get_exception()
        
        return response.result
        







#  WotoPlatform-Py - A Python library for interacting with WotoPlatform API.
#  Copyright (C) 2021-2022  ALiwoto - <woto@kaizoku.cyou> <https://github.com/ALiwoto>
#
#  This file is part of WotoPlatform-Py.
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from enum import Enum
from pydantic import BaseModel

class GeneralException(Exception):
    """
    General exception for the Woto platform.
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class ClientException(GeneralException):
    """
    client-side exceptions for the Woto platform.
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message

class ErrorCode(Enum):
    NoError = 0
    ErrUnknown = 1
    ErrServerUnavailable = 2
    ErrInvalidUsernameFormat = 3
    ErrInvalidPasswordFormat = 4
    ErrUsernameExists = 5
    ErrWrongUsername = 6
    ErrWrongPassword = 7
    ErrInvalidAuthKeyFormat = 8
    ErrInvalidAccessHashFormat = 9
    ErrWrongAuthKey = 10
    ErrLoginAccessHashExpired = 11
    ErrInvalidFirstName = 12
    ErrInvalidLastName = 13
    ErrInvalidTitle = 14
    ErrAlreadyAuthorized = 15
    ErrNotAuthorized = 16
    ErrNotModified = 17
    ErrBioTooLong = 18
    ErrUserNotFound = 19
    ErrFirstNameTooLong = 20
    ErrLastNameTooLong = 21
    ErrInvalidUsernameAndUserId = 22
    ErrMethodNotImplemented = 23
    ErrPermissionDenied = 24
    ErrKeyNotFound = 25
    ErrInvalidTelegramId = 26
    ErrInvalidEmail = 27

class ServerException(GeneralException):
    """
    server-side exceptions for the Woto platform.
    """
    code: ErrorCode = 0
    message: str = ''
    origin: str = ''
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return f"woto-platform says: [{self.code.name}] {self.message} Caused by {self.origin}."

class EndpointError(BaseModel):
    code: int
    message: str
    origin: str





#  WotoPlatform-Py - A Python library for interacting with WotoPlatform API.
#  Copyright (C) 2021-2022  ALiwoto - <woto@kaizoku.cyou> <https://github.com/ALiwoto>
#
#  This file is part of WotoPlatform-Py.
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import ssl
import sys
import socket
import time

try:
    import socks
except ImportError as e:
    e.msg = (
        "PySocks is missing and wotoplatform can't run without. "
        "Please install it using \"pip3 install pysocks\"."
    )

    raise e


class WotoSocket:
    TIMEOUT = 10
    
    _host: str = ''
    _port: int = 0
    _use_tls: bool = True
    _ipv6: bool = False
    _socket: socks.socksocket = None
    __reader: asyncio.StreamReader = None
    __writer: asyncio.StreamWriter = None
    conn_lock: asyncio.Lock = None
    is_initialized: bool = False
    is_closed: bool = False

    def __init__(
        self, 
        host: str, 
        port:int, 
        use_tls: bool, 
        ipv6: bool
    ) -> None:
        if not host:
            raise Exception('host is required')
        
        if not port:
            raise Exception('port is required')
        
        self.conn_lock = asyncio.Lock()
        self._host = host
        self._port = port
        self._use_tls = use_tls
        self._ipv6 = ipv6
        self.socket = socks.socksocket(
            socket.AF_INET6 if ipv6
            else socket.AF_INET
        )
        self.socket.settimeout(WotoSocket.TIMEOUT)
    
    async def connect(self, address: tuple = None):
        if address == None:
            address = (self._host, self._port)
        self.socket.connect(address)
        self.__reader, self.__writer = await asyncio.open_connection(sock=self.socket)
    
    async def connect_OLD(self, loop = None) -> None:
        if sys.platform == 'win32':
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
        ssl_ctx = None
        if self._use_tls:    
            ssl_ctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
            ssl_ctx.check_hostname = False
            ssl_ctx.set_ciphers('ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384')

        r, w = await self.open_connection(
            host=self._host, 
            port=self._port,
            loop=loop,
            ssl=ssl_ctx,
            ssl_handshake_timeout=9
        )
        self.__reader = r
        self.__writer = w
        self.is_initialized = True
    

    async def open_connection(self, host=None, port=None, loop=None, *,
                            limit=asyncio.streams._DEFAULT_LIMIT, **kwds):
        """A wrapper for create_connection() returning a (reader, writer) pair.

        The reader returned is a StreamReader instance; the writer is a
        StreamWriter instance.

        The arguments are all the usual arguments to create_connection()
        except protocol_factory; most common are positional host and port,
        with various optional keyword arguments following.

        Additional optional keyword arguments are loop (to set the event loop
        instance to use) and limit (to set the buffer limit passed to the
        StreamReader).

        (If you want to customize the StreamReader and/or
        StreamReaderProtocol classes, just copy the code -- there's
        really nothing special here except some convenience.)
        """
        if not loop:
            loop = asyncio.events.get_running_loop()
        reader = asyncio.StreamReader(limit=limit, loop=loop)
        protocol = asyncio.StreamReaderProtocol(reader, loop=loop)
        transport, _ = await loop.create_connection(
            lambda: protocol, host, port, **kwds)
        writer = asyncio.StreamWriter(transport, protocol, reader, loop)
        return reader, writer

    async def send(self, data: bytes) -> None:
        while not self.__writer and not self.is_initialized:
            await asyncio.sleep(0.1)
        
        async with self.conn_lock:
            self.__writer.write(data)
            await self.__writer.drain()
    
    async def recv(self, size: int) -> bytes:
        try:
            return await self.__reader.read(size)
        except IndexError:
            return b''
    
    async def close(self) -> None:
        try:
            try:
                self.__writer.close()
            except AttributeError:
                try:
                    self.socket.shutdown(socket.SHUT_RDWR)
                except OSError:
                    pass
                finally:
                    # A tiny sleep placed here helps avoiding .recv(n) hanging until the timeout.
                    # This is a workaround that seems to fix the occasional delayed stop of a client.
                    time.sleep(0.001)
                    self.socket.close()
            self.__reader = None
            self.__writer = None
            self.is_closed = True
        except Exception: pass
        
    


from pydantic import BaseModel
from typing import Iterable
import hashlib

class PasswordContainer256(BaseModel):
    """
    Contains a password from user input.
    Call `set_as_password` method on this class to set the user-input password.
    """
    __HEADERS_LEN = 0x03
    __HEADERS_CHARS = [0x2a, 0x21, 0x2f]
    __SIG_CHARS = [0x3c, 0x40, 0x5e, 0x2d, 0x7e, 0x5d, 0x3d, 0x29]
    
    PWD_ENCODING_VERSION = "<pwd-container::v1.0.0.0::256>"
    
    
    header: str = None
    hash256: str = None
    signature: str = None

    
    def set_as_password(self, text: str) -> None:
        self.calculate_header(text)
        self.calculate_hash(text)
        self.calculate_signature(text)

    def get_as_dict(self) -> dict:
        return {
            'header': self.header,
            'hash256': self.hash256,
            'signature': self.signature
        }
    
    def calculate_signature(self, user_input: str) -> None:
        #TODO: add support for payload data
        self.signature = self.__append_by_seps(user_input, self.__SIG_CHARS, True)
        if len(self.signature)%2 == 1:
            # See: https://github.com/TheGolangHub/wotoCrypto/blob/4877c7c9925871013917a8c300bef9bd2a73d4b5/wotoCrypto/passContainer/methods.go#L15
            self.signature += self.__SIG_CHARS[0x00]
    
    def calculate_hash(self, user_input: str) -> None:
        self.hash256 = hashlib.sha256(user_input.encode()).hexdigest()
    
    def calculate_header(self, user_input: str) -> None:
        headers_list = []
        headers_list.append(len(user_input)) # 0x00
        headers_list.append(0) # 0x01 (TODO: add support for payload in header)
        headers_list.append(self.PWD_ENCODING_VERSION)
        
        if not self.has_correct_headers_len(headers_list):
            raise Exception('Invalid headers len could get generated from the input value')
        
        self.header = self.__append_by_seps(headers_list, self.__HEADERS_CHARS)
    
    
    def has_correct_headers_len(self, headers: Iterable) -> bool:
        return len(headers) == self.__HEADERS_LEN
    
    
    def __append_by_seps(
        self, 
        slices: Iterable, 
        values: list, 
        to_num: bool = False
    ) -> str:
        result_value = ''
        current_index = -1
        for current in slices:
            if to_num: current = str(ord(current))
            current_index += 1
            result_value += str(current) + chr(values[current_index % len(values)])
        return result_value
    
    
    
    





#  WotoPlatform-Py - A Python library for interacting with WotoPlatform API.
#  Copyright (C) 2021-2022  ALiwoto - <woto@kaizoku.cyou> <https://github.com/ALiwoto>
#
#  This file is part of WotoPlatform-Py.
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from enum import Enum


class UserPermission(Enum):
    """
    User permissions.
    """
    NormalUser = 0
    Special = 1
    Admin = 2
    Developer = 3
    Owner = 4
    







#  WotoPlatform-Py - A Python library for interacting with WotoPlatform API.
#  Copyright (C) 2021-2022  ALiwoto - <woto@kaizoku.cyou> <https://github.com/ALiwoto>
#
#  This file is part of WotoPlatform-Py.
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .general import ServerException, ErrorCode


class UnknownError(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrUnknown
        self.message = message
        self.origin = origin
    

class ServerUnavailable(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrServerUnavailable
        self.message = message
        self.origin = origin


class InvalidUsernameFormat(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrInvalidUsernameFormat
        self.message = message
        self.origin = origin


class InvalidPasswordFormat(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrInvalidPasswordFormat
        self.message = message
        self.origin = origin


class UsernameExists(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrUsernameExists
        self.message = message
        self.origin = origin

class WrongUsername(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrWrongUsername
        self.message = message
        self.origin = origin

class WrongPassword(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrWrongPassword
        self.message = message
        self.origin = origin


class InvalidAuthKeyFormat(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrInvalidAuthKeyFormat
        self.message = message
        self.origin = origin


class InvalidAccessHashFormat(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrInvalidAccessHashFormat
        self.message = message
        self.origin = origin

class WrongAuthKey(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrWrongAuthKey
        self.message = message
        self.origin = origin

class LoginAccessHashExpired(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrLoginAccessHashExpired
        self.message = message
        self.origin = origin

class InvalidFirstName(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrInvalidFirstName
        self.message = message
        self.origin = origin


class InvalidLastName(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrInvalidLastName
        self.message = message
        self.origin = origin


class InvalidTitle(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrInvalidTitle
        self.message = message
        self.origin = origin


class AlreadyAuthorized(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrAlreadyAuthorized
        self.message = message
        self.origin = origin


class NotAuthorized(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrNotAuthorized
        self.message = message
        self.origin = origin

class NotModified(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrNotModified
        self.message = message
        self.origin = origin


class BioTooLong(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrBioTooLong
        self.message = message
        self.origin = origin

class UserNotFound(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrUserNotFound
        self.message = message
        self.origin = origin


class FirstNameTooLong(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrFirstNameTooLong
        self.message = message
        self.origin = origin

class LastNameTooLong(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrLastNameTooLong
        self.message = message
        self.origin = origin

class InvalidUsernameAndUserId(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrInvalidUsernameAndUserId
        self.message = message
        self.origin = origin

class MethodNotImplemented(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrMethodNotImplemented
        self.message = message
        self.origin = origin

class PermissionDenied(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrPermissionDenied
        self.message = message
        self.origin = origin

class KeyNotFound(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrKeyNotFound
        self.message = message
        self.origin = origin

class InvalidTelegramId(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrInvalidTelegramId
        self.message = message
        self.origin = origin

class InvalidEmail(ServerException):
    message: str = ''
    origin: str = ''

    def __init__(self, message: str, origin: str):
        self.code = ErrorCode.ErrInvalidEmail
        self.message = message
        self.origin = origin



#  WotoPlatform-Py - A Python library for interacting with WotoPlatform API.
#  Copyright (C) 2021-2022  ALiwoto - <woto@kaizoku.cyou> <https://github.com/ALiwoto>
#
#  This file is part of WotoPlatform-Py.
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .base_types.scaffold import *
from .data_types.version_data import *
from .data_types.users_data import *
from .clientBase import *
from .permissions import  *

__all__ = [
    RawResponse,
    RawDScaffold,
    ScaffoldHolder,
    Scaffold,
    DScaffold,
    EmptyScaffoldData,
    RScaffold,
    UserPermission,
    ClientBase,
    ResultScaffold,
    VersionData,
    VersionResponse,
    VersionResult,
    LoginUserData,
    LoginUserResponse,
    LoginUserResult,
    RegisterUserData,
    RegisterUserResponse,
    RegisterUserResult,
    ChangeUserBioData,
    ChangeUserBioResponse,
    GetUserFavoriteData,
    GetUserFavoriteResponse,
    GetUserFavoriteResult,
    GetUserFavoriteCountData,
    GetUserFavoriteCountResponse,
    GetUserFavoriteCountResult,
    SetUserFavoriteData,
    SetUserFavoriteResponse,
    ResolveUsernameData,
    ResolveUsernameResponse,
    ResolveUsernameResult,
]



#  WotoPlatform-Py - A Python library for interacting with WotoPlatform API.
#  Copyright (C) 2021-2022  ALiwoto - <woto@kaizoku.cyou> <https://github.com/ALiwoto>
#
#  This file is part of WotoPlatform-Py.
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional

from wotoplatform.types.woto_crypto.password_container import PasswordContainer256

from .. import (
    DScaffold,
    RScaffold,
    ResultScaffold,
    EmptyScaffoldData,
)
from ..permissions import UserPermission



__BATCH_REGISTER_USER__             = 'register_user'
__BATCH_LOGIN_USER__                = 'login_user'
__BATCH_GET_ME__                    = 'get_me'
__BATCH_CHANGE_USER_BIO__           = 'change_user_bio'
__BATCH_CHANGE_NAMES__              = 'change_names'
__BATCH_CHANGE_USER_INFO__          = 'change_user_info'
__BATCH_GET_USER_INFO__             = 'get_user_info'
__BATCH_RESOLVE_USERNAME__          = 'resolve_username'
__BATCH_GET_USER_FAVORITE__         = 'get_user_favorite'
__BATCH_GET_USER_FAVORITE_COUNT__   = 'get_user_favorite_count'
__BATCH_SET_USER_FAVORITE__         = 'set_user_favorite'
__BATCH_DELETE_USER_FAVORITE__      = 'delete_user_favorite'


__ACTION_USER__ = 2

class RegisterUserData(DScaffold):
    user_id: int = 0
    private_hash: str = ''
    username: str = ''
    password: PasswordContainer256 = None
    first_name: str = ''
    last_name: str = ''
    email: str = ''
    permission: UserPermission = UserPermission.NormalUser

    def get_action(self) -> int:
        return __ACTION_USER__
    
    def get_single_batch(self) -> str:
        return __BATCH_REGISTER_USER__
    
    def get_response_type(self) -> type:
        return RegisterUserResponse
    
class RegisterUserResult(ResultScaffold):
    user_id: int = 0
    private_hash: str = ''
    email: str = ''
    website: str = ''
    auth_key: str = ''
    access_hash: str = ''
    permission: UserPermission = UserPermission.NormalUser
    bio: str = ''
    source_url: str = ''
    telegram_id: int = 0
    first_name: str = ''
    last_name: str = ''
    username: str = ''
    created_at: str = ''
    updated_at: str = ''
    is_virtual: bool = False
    created_by: int = 0
    
class RegisterUserResponse(RScaffold):
    result: Optional[RegisterUserResult]


class LoginUserData(DScaffold):
    username: str = ''
    password: PasswordContainer256 = None
    auth_key: str = ''
    access_hash: str = ''

    def get_action(self) -> int:
        return __ACTION_USER__
    
    def get_single_batch(self) -> str:
        return __BATCH_LOGIN_USER__
    
    def get_response_type(self) -> type:
        return LoginUserResponse

class LoginUserResult(ResultScaffold):
    user_id: int = 0
    private_hash: str = ''
    email: str = ''
    website: str = ''
    permission: UserPermission = UserPermission.NormalUser
    bio: str = ''
    source_url: str = ''
    telegram_id: int = 0
    first_name: str = ''
    last_name: str = ''
    username: str = ''
    created_at: str = ''
    updated_at: str = ''
    is_virtual: bool = False
    created_by: int = 0
    
class LoginUserResponse(RScaffold):
    result: Optional[LoginUserResult]


class GetMeData(EmptyScaffoldData):
    def get_action(self) -> int:
        return __ACTION_USER__
    
    def get_single_batch(self) -> str:
        return __BATCH_GET_ME__
    
    def get_response_type(self) -> type:
        return GetMeResponse

class GetMeResult(ResultScaffold):
    user_id: int = 0
    private_hash: str = ''
    email: str = ''
    website: str = ''
    permission: UserPermission = UserPermission.NormalUser
    bio: str = ''
    source_url: str = ''
    telegram_id: int = 0
    first_name: str = ''
    last_name: str = ''
    username: str = ''
    created_at: str = ''
    updated_at: str = ''
    is_virtual: bool = False
    created_by: int = 0
    
class GetMeResponse(RScaffold):
    result: Optional[LoginUserResult]


class ChangeUserBioData(DScaffold):
    user_id: int = 0
    bio: str = ''

    def get_action(self) -> int:
        return __ACTION_USER__
    
    def get_single_batch(self) -> str:
        return __BATCH_CHANGE_USER_BIO__
    
    def get_response_type(self) -> type:
        return ChangeUserBioResponse

class ChangeUserBioResponse(RScaffold):
    result: Optional[bool]


class ChangeNamesData(DScaffold):
    user_id: int = 0
    first_name: str = ''
    last_name: str = ''

    def get_action(self) -> int:
        return __ACTION_USER__
    
    def get_single_batch(self) -> str:
        return __BATCH_CHANGE_NAMES__
    
    def get_response_type(self) -> type:
        return ChangeUserBioResponse

class ChangeUserBioResponse(RScaffold):
    result: Optional[bool]


class GetUserInfoData(DScaffold):
    user_id: int = 0
    username: str = ''

    def get_action(self) -> int:
        return __ACTION_USER__
    
    def get_single_batch(self) -> str:
        return __BATCH_GET_USER_INFO__
    
    def get_response_type(self) -> type:
        return GetUserInfoResponse

class GetUserInfoResult(ResultScaffold):
    user_id: int = 0
    email: str = ''
    region: str = ''
    language: str = ''
    birthday: str = ''
    anilist_url: str = ''
    my_animelist_url: str = ''
    reddit_url: str = ''
    github_url: str = ''
    gitlab_url: str = ''
    favorite_color: str = ''
    favorite_music: str = ''
    favorite_anime: str = ''
    favorite_movie: str = ''
    favorite_food: str = ''
    favorite_series: str = ''
    favorite_light_novel: str = ''
    favorite_novel: str = ''
    website: str = ''
    permission: UserPermission = UserPermission.NormalUser
    bio: str = ''
    source_url: str = ''
    telegram_id: int = 0
    first_name: str = ''
    last_name: str = ''
    username: str = ''
    created_at: str = ''
    updated_at: str = ''
    is_virtual: bool = False
    
class GetUserInfoResponse(RScaffold):
    result: Optional[GetUserInfoResult]


class ResolveUsernameData(DScaffold):
    username: str = ''

    def get_action(self) -> int:
        return __ACTION_USER__
    
    def get_single_batch(self) -> str:
        return __BATCH_LOGIN_USER__
    
    def get_response_type(self) -> type:
        return ResolveUsernameResponse

class ResolveUsernameResult(ResultScaffold):
    is_acceptable: bool = False
    server_time: str = ''
    
class ResolveUsernameResponse(RScaffold):
    result: Optional[LoginUserResult]


class GetUserFavoriteData(DScaffold):
    user_id: int = 0
    favorite_key: str = ''

    def get_action(self) -> int:
        return __ACTION_USER__
    
    def get_single_batch(self) -> str:
        return __BATCH_GET_USER_FAVORITE__
    
    def get_response_type(self) -> type:
        return GetUserFavoriteResponse


class GetUserFavoriteResult(ResultScaffold):
    favorite_value: str = ''
    updated_at: str = ''

class GetUserFavoriteResponse(RScaffold):
    result: Optional[GetUserFavoriteResult]


class GetUserFavoriteCountData(DScaffold):
    user_id: int = 0

    def get_action(self) -> int:
        return __ACTION_USER__
    
    def get_single_batch(self) -> str:
        return __BATCH_GET_USER_FAVORITE_COUNT__
    
    def get_response_type(self) -> type:
        return GetUserFavoriteCountResponse


class GetUserFavoriteCountResult(ResultScaffold):
    favorites_count: int = 0

class GetUserFavoriteCountResponse(RScaffold):
    result: Optional[GetUserFavoriteCountResult]

class SetUserFavoriteData(DScaffold):
    user_id: int = 0
    favorite_key: str = ''
    favorite_value: str = ''

    def get_action(self) -> int:
        return __ACTION_USER__
    
    def get_single_batch(self) -> str:
        return __BATCH_SET_USER_FAVORITE__
    
    def get_response_type(self) -> type:
        return SetUserFavoriteResponse


class SetUserFavoriteResponse(RScaffold):
    result: Optional[bool]

class DeleteUserFavoriteData(DScaffold):
    user_id: int = 0
    favorite_key: str = ''

    def get_action(self) -> int:
        return __ACTION_USER__
    
    def get_single_batch(self) -> str:
        return __BATCH_DELETE_USER_FAVORITE__
    
    def get_response_type(self) -> type:
        return DeleteUserFavoriteResponse

class DeleteUserFavoriteResponse(RScaffold):
    result: Optional[bool]

class ResolveUsernameData(DScaffold):
    username: str = ''

    def get_action(self) -> int:
        return __ACTION_USER__
    
    def get_single_batch(self) -> str:
        return __BATCH_DELETE_USER_FAVORITE__
    
    def get_response_type(self) -> type:
        return ResolveUsernameResponse

class ResolveUsernameResponse(RScaffold):
    result: Optional[GetUserInfoResult]


class LikedItem(DScaffold):
    """
    type LikedListElement struct {
        UniqueId     string       `json:"unique_id" gorm:"primaryKey"`
        OwnerId      PublicUserId `json:"owner_id"`
        MediaId      MediaModelId `json:"media_id"`
        Title        string       `json:"title"`
        LikedKey     string       `json:"liked_key"`
        SourceUrl    string       `json:"source_url"`
        ReferenceUrl string       `json:"reference_url"`
        UpdatedAt    time.Time    `json:"-"`
    }
    """
    unique_id: str = ''
    owner_id: int = 0
    media_id: int = 0
    title: str = ''
    liked_key: str = ''
    source_url: str = ''
    reference_url: str = ''





#  WotoPlatform-Py - A Python library for interacting with WotoPlatform API.
#  Copyright (C) 2021-2022  ALiwoto - <woto@kaizoku.cyou> <https://github.com/ALiwoto>
#
#  This file is part of WotoPlatform-Py.
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import typing
from ..base_types.scaffold import (
    DScaffold,
    RScaffold,
    ResultScaffold,
)

__BATCH_CHECK_VERSION__ = 'check_version'
__ACTION_VERSION__ = 1

class VersionData(DScaffold):
    user_agent: str = 'wp-client'
    version_key: str = '2.1.1.5014'
    version_hash: str = 'f302bd7ffacbd295194f86620002b8250e8e9be0233a8055bcebc82c8612843ff9e0f09e42015d5e75581cc93d4c29a91388ed411641b543c8fb7b5a26a2a8b8'
    client_id: str = 'cli-12345678910'

    def get_action(self) -> int:
        return __ACTION_VERSION__
    
    def get_single_batch(self) -> str:
        return __BATCH_CHECK_VERSION__
    
    def get_response_type(self) -> type:
        return VersionResponse


class VersionResult(ResultScaffold):
    is_acceptable: bool = False
    server_time: str = ''
    
class VersionResponse(RScaffold):
    result: typing.Optional[VersionResult]




#  WotoPlatform-Py - A Python library for interacting with WotoPlatform API.
#  Copyright (C) 2021-2022  ALiwoto - <woto@kaizoku.cyou> <https://github.com/ALiwoto>
#
#  This file is part of WotoPlatform-Py.
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
import typing
from enum import Enum
from pydantic import BaseModel


__CR_LAYER_FACTOR__ = 0x1b
__BA_ALGORITHM_FACTOR__ = 0x2

class WotoAlgorithm(Enum):
    M250 = 0x40
    M251 = 0x30
    M252 = 0x28
    M253 = 0x20

class BlockAlgorithmId(Enum):
    X917 = __BA_ALGORITHM_FACTOR__ << 0x0
    X847 = __BA_ALGORITHM_FACTOR__ << 0x1
    X795 = __BA_ALGORITHM_FACTOR__ << 0x2
    X649 = __BA_ALGORITHM_FACTOR__ << 0x3


class CryptoLayerKind(Enum):
    O27 = __CR_LAYER_FACTOR__ << 0x0
    O54 = __CR_LAYER_FACTOR__ << 0x1
    O108 = __CR_LAYER_FACTOR__ << 0x2
    O216 = __CR_LAYER_FACTOR__ << 0x3



class WotoLayerLength(int):
    pass

class _PrivateBlock(int):
    pass


__layer_kinds_map__ = {
    CryptoLayerKind.O27: True,
    CryptoLayerKind.O54: True,
    CryptoLayerKind.O108: True,
    CryptoLayerKind.O216: True,
}

__layer_length_validator = {
    CryptoLayerKind.O27: lambda x: x == 0x27,
    CryptoLayerKind.O54: lambda x: x == 0x54,
    CryptoLayerKind.O108: lambda x: x == 0x108,
    CryptoLayerKind.O216: lambda x: x == 0x216,
}

class WotoValidator():
    def is_valid(self) -> bool:
        pass


class WotoSerializer(BaseModel):
    def serialize_object(self) -> bytes:
        pass

    def str_serialize(self) -> str:
        pass


class BytesObject():
    def to_bytes(self) -> bytes:
        pass

    def get_length(self) -> int:
        pass

def __encrypt_data__(key: bytes, data: bytes) -> bytes:
    key_blocks = __to_block_collection(key)
    data_blocks = __to_block_collection(data)
    return __encrypt_by_blocks(key_blocks, data_blocks, None)

def __decrypt_data__(key: bytes, data: bytes) -> bytes:
    key_blocks = __to_block_collection(key)
    data_blocks = __to_block_collection(data)
    return __decrypt_by_blocks(key_blocks, data_blocks, None)

def __encrypt_by_blocks(key: '_BlockCollection', data: '_BlockCollection', algorithm: '_BlockAlgorithm') -> bytes:
    pass

def __decrypt_by_blocks(key: '_BlockCollection', data: '_BlockCollection', algorithm: '_BlockAlgorithm') -> bytes:
    pass

def __to_block_collection(data: bytes) -> '_BlockCollection':
    return _PrivateCollection(data)

class _PrivateCollection:
    blocks: typing.List[_PrivateBlock]

    def __init__(self, data: bytes):
        my_data = data.decode('utf-8')
        self.blocks = []
        for current in my_data:
            self.blocks.append(_PrivateBlock(ord(current)))

class LayerLengthContainer(BaseModel, WotoValidator):
    length: WotoLayerLength = 0
    layer_kind: CryptoLayerKind = CryptoLayerKind.O27

    def is_valid(self) -> bool:
        return self.length > 0

    

class CryptoLayer(BaseModel, WotoValidator):
    kind: CryptoLayerKind
    hash: str
    lenContainer: typing.Optional[LayerLengthContainer] = None

    def get_layer_length(self) -> LayerLengthContainer:
        if self.lenContainer is not None and self.lenContainer.is_valid():
            return self.lenContainer
        self.lenContainer = self.get_new_layer_container()
        return self.lenContainer
    
    def to_map(self):
        return {
            'kind': self.kind.value,
            'hash': self.hash
        }
    
    def get_new_layer_container(self) -> LayerLengthContainer:
        return LayerLengthContainer(length=self.get_length(), layer_kind=self.kind)
    
    def get_length(self) -> WotoLayerLength:
        return WotoLayerLength(len(self.hash))
    
    def is_valid(self) -> bool:
        return self.get_layer_length().is_valid()
    
    def to_bytes(self) -> bytes:
        if not self.is_valid():
            raise Exception("Invalid CryptoLayer")
        return bytes(self.hash, encoding='utf-8')
    
    def is_equal(self, layer: 'CryptoLayer') -> bool:
        return self.hash == layer.hash and self.kind == layer.kind

class KeyLayerCollection(typing.List[CryptoLayer]):
    def get_layer_by_index(self, index: int) -> CryptoLayer:
        if index >= len(self):
            return None
        return self[index]
    
    def is_valid(self) -> bool:
        return len(self) != 0 and self.validate_keys()
    
    def validate_keys(self) -> bool:
        for current in self:
            if not current.is_valid():
                return False
        return True

    def to_list(self):
        return [current.to_map() for current in self]
    
    def contains(self, layer: CryptoLayer) -> bool:
        for current in self:
            if current.is_equal(layer):
                return True
        return False
    
    def contains_kind(self, kind: CryptoLayerKind) -> bool:
        for current in self:
            if current.kind == kind:
                return True
        return False
    
    def get_key_length(self) -> int:
        total = 0
        for current in self:
            total += int(current.get_length())
        return total
    

    def get_layer_length_by_kind(self, kind: CryptoLayerKind) -> LayerLengthContainer:
        for current in self:
            if current.kind == kind:
                return current.get_layer_length()
        return None

class _BlockAlgorithmX917(BaseModel):
    identifier: int


class _BlockAlgorithmX847(BaseModel):
    identifier: int


class _BlockAlgorithmX795(BaseModel):
    identifier: int


class _BlockAlgorithmX649(BaseModel):
    identifier: int


class AlgorithmSupporter(BaseModel):
    """
    SetAlgorithm(algorithm WotoAlgorithm) bool
	HasEqualAlgorithm(algorithm WotoAlgorithm) bool
	GetAlgorithm() WotoAlgorithm
	GetHashCount() int
    """
    def set_algorithm(self, algorithm: WotoAlgorithm) -> bool:
        pass

    def has_equal_algorithm(self, algorithm: WotoAlgorithm) -> bool:
        pass

    def get_algorithm(self) -> WotoAlgorithm:
        pass

    def get_gash_count(self) -> int:
        pass


class LayerBlock(BaseModel):
    """
    ContainsLayerKind(kind CryptoLayerKind) bool
	ContainsLayer(layer *CryptoLayer) bool
	AppendLayer(layer *CryptoLayer) bool
	RemoveLayer(layer *CryptoLayer) bool
	RemoveLayers(layers ...*CryptoLayer)
	GetLayerLengthByKind(kind CryptoLayerKind) *LayerLengthContainer
	GetLayerLengthByIndex(index int) *LayerLengthContainer
	GetKeyLayersCount() int
    """
    def contains_layer_kind(self, kind: CryptoLayerKind) -> bool:
        pass

    def contains_layer(self, layer: CryptoLayer) -> bool:
        pass

    def append_layer(self, layer: CryptoLayer) -> bool:
        pass

    def remove_layer(self, layer: CryptoLayer) -> bool:
        pass

    def remove_layers(self, layers: typing.List[CryptoLayer]):
        pass

    def get_layer_length_by_kind(self, kind: CryptoLayerKind) -> LayerLengthContainer:
        pass

    def get_layer_length_by_index(self, index: int) -> LayerLengthContainer:
        pass

    def get_key_layers_count(self) -> int:
        pass


class SignatureContainer():
    def get_signature(self) -> str:
        pass

    def set_signature(self, signature: str) -> bool:
        pass

    def set_signature_by_bytes(self, signature: bytes) -> bool:
        if not signature:
            return False
        return self.set_signature(signature.decode('utf-8'))


class IntegerRepresent():
    def to_int(self) -> int:
        pass


class BitsBlocks():
    def get_bits_size() -> int:
        pass

class WotoKey(WotoValidator, LayerBlock, AlgorithmSupporter, WotoSerializer, SignatureContainer):
    """
	IsFuture() bool
	IsPast() bool
	IsPresent() bool
	IsEmpty() bool
	CanBecomeFuture() bool
	CanBecomePresent() bool
	CanBecomePast() bool
	Decrypt(data []byte) []byte
	Encrypt(data []byte) []byte
	HasEqualKind(key WotoKey) bool
	HasEqualSignature(key WotoKey) bool
	GetKeyLength() int
	GetSignatureRealLength() int
	IsRealLengthInvalid() bool

	// Deprecated: you can't convert any WotoKey to a FutureKey anymore.
	// Please use GenerateFutureKey helper function.
	ToFutureKey() WotoKey
	ToPresentKey() WotoKey
	ToPastKey() WotoKey
	Clone() WotoKey

	getLayers() KeyLayerCollection
	setLayers(layers KeyLayerCollection) bool
    """

    def is_future(self) -> bool:
        pass

    def is_past(self) -> bool:
        pass

    def is_present(self) -> bool:
        pass

    def is_empty(self) -> bool:
        pass

    def can_become_future(self) -> bool:
        pass

    def can_become_present(self) -> bool:
        pass

    def can_become_past(self) -> bool:
        pass

    def decrypt_data(self, data: bytes) -> bytes:
        pass

    def encrypt_data(self, data: bytes) -> bytes:
        pass

    def has_equal_kind(self, key: 'WotoKey') -> bool:
        pass

    def has_equal_signature(self, key: 'WotoKey') -> bool:
        pass

    def get_key_length(self) -> int:
        pass

    def get_signature_real_length(self) -> int:
        pass

    def is_real_length_invalid(self) -> bool:
        pass

    def to_future_key(self) -> 'WotoKey':
        pass

    def to_present_key(self) -> 'WotoKey':
        pass

    def to_past_key(self) -> 'WotoKey':
        pass

    def clone(self) -> 'WotoKey':
        pass

    def _get_layers(self) -> typing.List[CryptoLayer]:
        pass

    def _set_layers(self, layers: typing.List[CryptoLayer]) -> bool:
        pass

    


class KeyCollection(BaseModel, WotoValidator):
    def continue_life_cycle(self):
        pass

    def sync(self):
        pass


class KeysContainer():
    def set_as_keys(self, value: KeyCollection):
        pass


class _SingleBlock(WotoValidator, BitsBlocks):
    """
    IsEmpty() bool
	IsNonZero() bool
	ToInt64() int64
	ToUInt64() uint64
	ToInt32() int32
	ToUInt32() uint32
	Sum(singleBlock) singleBlock
	Min(singleBlock) singleBlock
	Mul(singleBlock) singleBlock
	Div(singleBlock) singleBlock
    """

    def is_empty(self) -> bool:
        pass

    def is_non_zero(self) -> bool:
        pass

    def to_int64(self) -> int:
        pass

    def to_uint64(self) -> int:
        pass

    def to_int32(self) -> int:
        pass

    def to_uint32(self) -> int:
        pass

    def sum(self, singleBlock: '_SingleBlock') -> '_SingleBlock':
        pass

    def min(self, singleBlock: '_SingleBlock') -> '_SingleBlock':
        pass

    def mul(self, singleBlock: '_SingleBlock') -> '_SingleBlock':
        pass

    def div(self, singleBlock: '_SingleBlock') -> '_SingleBlock':
        pass


# type blockAction func(first, second singleBlock) singleBlock

class _BlockAction(typing.Callable[[_SingleBlock, _SingleBlock], _SingleBlock]):
    pass

class _BlockCollection(BytesObject):
    """
    GetBlocks() []singleBlock
	GetRelativeIndex(int) int
	BlockSize() int
	AppendBlock(singleBlock)
	AppendCollection(blockCollection)
	GetBlockByIndex(int) singleBlock
	Clone() blockCollection
    """

    def get_blocks(self) -> typing.List[_SingleBlock]:
        pass

    def get_relative_index(self, index: int) -> int:
        pass

    def block_size(self) -> int:
        pass

    def append_block(self, singleBlock: _SingleBlock):
        pass

    def append_collection(self, blockCollection: '_BlockCollection'):
        pass

    def get_block_by_index(self, index: int) -> _SingleBlock:
        pass

    def clone(self) -> '_BlockCollection':
        pass

#2028391795 https://t.me/SpaceXSupports

class _BlockAlgorithm():
    """
    GetEncryptBlockAction(index int) blockAction
	GetDecryptBlockAction(index int) blockAction
    """

    def get_encrypt_block_action(self, index: int) -> _BlockAction:
        pass

    def get_decrypt_block_action(self, index: int) -> _BlockAction:
        pass



class _FutureKey(WotoKey):
    key_layers: KeyLayerCollection
    algorithm: WotoAlgorithm
    sig: str

    def to_future_key(self) -> 'WotoKey':
        return self
    
    def to_present_key(self) -> 'WotoKey':
        return _PresentKey(self.key_layers, self.algorithm, self.sig)
    
    def to_past_key(self) -> 'WotoKey':
        return None


class _PresentKey(WotoKey):
    key_layers: KeyLayerCollection
    algorithm: WotoAlgorithm
    sig: str
    def __init__(self, keyLayers: 'KeyLayerCollection', algorithm: 'WotoAlgorithm', sig: str):
        self.key_layers = keyLayers
        self.algorithm = algorithm
        self.sig = sig

    

    """

func (p *presentKey) GetLayers() KeyLayerCollection {
	return p.keyLayers
}

func (p *presentKey) GetLayerLengthByIndex(index int) *LayerLengthContainer {
	return p.keyLayers.GetLayerByIndex(index).GetLayerLength()
}

func (p *presentKey) SetLayers(layers KeyLayerCollection) bool {
	if !layers.IsValid() || !p.isValidWithAlgo(layers) {
		return false
	}

	p.keyLayers = layers

	return true
}

func (p *presentKey) isValidWithAlgo(layers KeyLayerCollection) bool {
	return true
}

func (p *presentKey) SetAlgorithm(algorithm WotoAlgorithm) bool {
	p.algorithm = algorithm
	return true
}

func (p *presentKey) GetSignature() string {
	return p.sig
}

func (p *presentKey) IsValid() bool {
	return p != nil && !p.IsEmpty() && p.sig != ""
}

func (p *presentKey) IsEmpty() bool {
	return len(p.keyLayers) == 0x0
}

func (p *presentKey) SetSignature(signature string) bool {
	if signature == "" {
		return false
	}

	p.sig = signature
	return true
}

func (p *presentKey) SetSignatureByBytes(data []byte) bool {
	if len(data) == 0 {
		return false
	}
	return p.SetSignature(string(data))
}

func (p *presentKey) SetSignatureByFunc(h func() hash.Hash) bool {
	if h == nil {
		return false
	}
	return p.SetSignatureByBytes(h().Sum(nil))
}

func (p *presentKey) Encrypt(data []byte) []byte {
	if !p.IsValid() {
		return data
	}

	switch p.algorithm {
	case WotoAlgorithmM250:
		return p.encryptM250(data)
	}
	return nil
}

func (p *presentKey) Serialize() ([]byte, error) {
	if !p.IsValid() {
		return nil, ErrInvalidKey
	}

	b, err := json.Marshal(p.toMap())
	if err != nil {
		return nil, err
	}

	return b, nil
}

func (p *presentKey) StrSerialize() string {
	b, err := p.Serialize()
	if err != nil || len(b) == 0 {
		return ""
	}

	return string(b)
}

func (p *presentKey) toMap() map[string]interface{} {
	return map[string]interface{}{
		"key_layers": p.keyLayers,
		"signature":  p.sig,
		"algorithm":  p.algorithm,
	}
}

func (p *presentKey) Decrypt(data []byte) []byte {
	if !p.IsValid() {
		return data
	}

	switch p.algorithm {
	case WotoAlgorithmM250:
		return p.decryptM250(data)
	}
	return nil
}

func (p *presentKey) encryptM250(data []byte) []byte {
	currentKey, err := p.keyLayers[0x0].ToBytes()
	if err != nil {
		return nil
	}
	var buff []byte

	for i, currentLayer := range p.keyLayers {
		if i == 0x0 {
			continue
		}

		buff, err = currentLayer.ToBytes()
		if err != nil {
			return nil
		}

		currentKey = EncryptData(currentKey, buff)
	}

	return EncryptData(currentKey, data)
}

func (p *presentKey) decryptM250(data []byte) []byte {
	currentKey, err := p.keyLayers[0x0].ToBytes()
	if err != nil {
		return nil
	}

	var buff []byte
	for i, currentLayer := range p.keyLayers {
		if i == 0x0 {
			continue
		}

		buff, err = currentLayer.ToBytes()
		if err != nil {
			return nil
		}

		currentKey = EncryptData(currentKey, buff)
	}

	return DecryptData(currentKey, data)
}

func (p *presentKey) AppendLayer(layer *CryptoLayer) bool {
	if !layer.IsValid() {
		return false
	}

	p.keyLayers = append(p.keyLayers, *layer)
	return true
}

func (p *presentKey) RemoveLayer(layer *CryptoLayer) bool {
	var newLayers KeyLayerCollection
	var done bool
	for _, current := range p.keyLayers {
		if !done && current.Equal(layer) {
			continue
		}
		newLayers = append(newLayers, current)
	}

	p.keyLayers = newLayers
	return true
}

func (p *presentKey) CanBecomeFuture() bool {
	return false
}

func (p *presentKey) CanBecomePast() bool {
	return p.algorithm&0x25 != 0x78
}

func (p *presentKey) CanBecomePresent() bool {
	return true
}

func (p *presentKey) ContainsLayer(layer *CryptoLayer) bool {
	if len(p.keyLayers) == 0x0 {
		return false
	}
	return p.keyLayers.Contains(layer)
}

func (p *presentKey) ContainsLayerKind(kind CryptoLayerKind) bool {
	if len(p.keyLayers) == 0x0 {
		return false
	}
	return p.keyLayers.ContainsKind(kind)
}

func (p *presentKey) GetAlgorithm() WotoAlgorithm {
	return p.algorithm
}

func (p *presentKey) GetHashCount() int {
	return len(p.keyLayers)
}

func (p *presentKey) GetKeyLayersCount() int {
	return len(p.keyLayers)
}

func (p *presentKey) GetKeyLength() int {
	if len(p.keyLayers) == 0x0 {
		return 0x0
	}
	return p.keyLayers.GetKeyLength()
}

func (p *presentKey) GetLayerLengthByKind(kind CryptoLayerKind) *LayerLengthContainer {
	if len(p.keyLayers) == 0x0 {
		return nil
	}
	return p.keyLayers.GetLayerLengthByKind(kind)
}

func (p *presentKey) HasEqualAlgorithm(algorithm WotoAlgorithm) bool {
	return p.algorithm == algorithm
}

func (p *presentKey) HasEqualKind(key WotoKey) bool {
	return key.IsPresent()
}

func (p *presentKey) HasEqualSignature(key WotoKey) bool {
	return p.sig == key.GetSignature()
}

func (p *presentKey) IsFuture() bool {
	return false
}

func (p *presentKey) IsPast() bool {
	return false
}

func (p *presentKey) IsPresent() bool {
	return true
}

func (p *presentKey) RemoveLayers(layers ...*CryptoLayer) {
	for _, layer := range layers {
		p.RemoveLayer(layer)
	}
}

func (p *presentKey) GetSignatureRealLength() int {
	if p.sig == "" {
		return 0x0
	}
	myStr := C.CString(p.sig)
	defer C.free(unsafe.Pointer(myStr))
	return int(C.compute_signature_real_length(myStr, C.int(p.algorithm)))
}

func (p *presentKey) IsRealLengthInvalid() bool {
	if p.sig == "" {
		return true
	}

	myStr := C.CString(p.sig)
	myL := C.int(p.algorithm)
	defer C.free(unsafe.Pointer(myStr))
	return C.is_real_length_invalid(C.compute_signature_real_length(myStr, myL), myL) > 0
}

func (p *presentKey) String() string {
	return "PresentKey with signature: " + p.sig
}

func (p *presentKey) ToFutureKey() WotoKey {
	return &futureKey{
		keyLayers: p.keyLayers,
		algorithm: p.algorithm,
		sig:       p.sig,
	}
}

func (p *presentKey) Clone() WotoKey {
	return &presentKey{
		keyLayers: p.keyLayers,
		algorithm: p.algorithm,
		sig:       p.sig,
	}
}

func (p *presentKey) ToPastKey() WotoKey {
	return &pastKey{
		keyLayers: p.keyLayers,
		algorithm: p.algorithm,
		sig:       p.sig,
	}
}

func (p *presentKey) ToPresentKey() WotoKey {
	return p
}

func (p *presentKey) getLayers() KeyLayerCollection {
	return p.keyLayers
}

func (p *presentKey) setLayers(layers KeyLayerCollection) bool {
	p.keyLayers = layers
	return true
}
    """

    def _get_layers(self) -> KeyLayerCollection:
        return self.key_layers
    
    def get_layer_length_by_index(self, index: int) -> LayerLengthContainer:
        if index < 0 or index >= len(self.key_layers):
            raise IndexError("Index out of range")
        return self.key_layers[index].get_length()

    def _set_layers(self, layers: KeyLayerCollection) -> bool:
        if not layers.is_valid() or not self.is_valid_with_algo(layers):
            return False
        self.key_layers = layers
        return True
    
    def is_valid_with_algo(self, layers: KeyLayerCollection) -> bool:
        return True # TODO: Implement
    
    def set_algorithm(self, algorithm: WotoAlgorithm) -> bool:
        self.algorithm = algorithm
        return True
    
    def get_signature(self) -> str:
        return self.sig
    
    def is_valid(self) -> bool:
        return bool(not self.is_empty() and self.sig)

    def is_empty(self) -> bool:
        return len(self.key_layers) == 0
    
    def set_signature(self, signature: str) -> bool:
        if not signature:
            return False
        self.sig = signature
        match self.algorithm:
            case WotoAlgorithm.M251:
                return
        return True
    
    def encrypt_data(self, data: bytes) -> bytes:
        if not self.is_valid():
            return data
        if self.algorithm == WotoAlgorithm.M250:
            return self.__encrypt_m250(data)
        elif self.algorithm == WotoAlgorithm.M251:
            return self.__encrypt_m251(data)
        elif self.algorithm == WotoAlgorithm.M252:
            return self.__encrypt_m252(data)
        elif self.algorithm == WotoAlgorithm.M253:
            return self.__encrypt_m253(data)
        else:
            return data # unknown algorithm
        
    
    def __encrypt_m250(self, data: bytes) -> bytes:
        current_key = self.key_layers[0x0].to_bytes()
        buf: bytes = None

        for i in range(0x0, len(self.key_layers)):
            buf = self.key_layers[i].to_bytes(current_key)
            current_key = __encrypt_data__(current_key, buf)
        
        return __encrypt_data__(current_key, data)



    def __encrypt_m251(self, data: bytes) -> bytes:
        pass


    def __encrypt_m252(self, data: bytes) -> bytes:
        pass


    def __encrypt_m253(self, data: bytes) -> bytes:
        pass

    def serialize_object(self) -> bytes:
        if not self.is_valid():
            raise ValueError("Invalid key")
        return json.dumps(self.to_map()).encode()
    
    def to_map(self):
        return {
            "algorithm": self.algorithm.value,
            "key_layers": self.key_layers.to_list(),
            "signature": self.sig
        }
    
    def str_serialize(self) -> str:
        return json.dumps(self.to_map())
    
    
    def decrypt_data(self, data: bytes) -> bytes:
        if self.algorithm == WotoAlgorithm.M250:
            return self.__decrypt_m250(data)
        elif self.algorithm == WotoAlgorithm.M251:
            return self.__decrypt_m251(data)
        elif self.algorithm == WotoAlgorithm.M252:
            return self.__decrypt_m252(data)
        elif self.algorithm == WotoAlgorithm.M253:
            return self.__decrypt_m253(data)
        else:
            return data # unknown algorithm
        
    
    def __decrypt_m250(self, data: bytes) -> bytes:
        current_key = self.key_layers[0x0].to_bytes()
        buf: bytes = None

        for i in range(0x0, len(self.key_layers)):
            buf = self.key_layers[i].to_bytes(current_key)
            current_key = __encrypt_data__(current_key, buf)
        
        return __decrypt_data__(current_key, data)


    def __decrypt_m251(self, data: bytes) -> bytes:
        pass


    def __decrypt_m252(self, data: bytes) -> bytes:
        pass


    def __decrypt_m253(self, data: bytes) -> bytes:
        pass
    
    
    


class _PastKey(WotoKey):
    key_layers: KeyLayerCollection
    algorithm: WotoAlgorithm
    sig: str

    def to_future_key(self) -> 'WotoKey':
        return None
    
    def to_present_key(self) -> 'WotoKey':
        return None
    
    def to_past_key(self) -> 'WotoKey':
        return self


class WotoKeyCollection():
    
    def parse_collection(value: str) -> 'WotoKeyCollection':
        json.load(value)
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wotoplatform',
 'wotoplatform.types',
 'wotoplatform.types.base_types',
 'wotoplatform.types.data_types',
 'wotoplatform.types.errors',
 'wotoplatform.utils',
 'wotoplatform.utils.woto_crypto']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.8.2', 'pysocks>=1.7.1']

setup_kwargs = {
    'name': 'wotoplatform',
    'version': '0.0.21',
    'description': '',
    'long_description': '\n### python library for woto-platform\n',
    'author': 'AliWoto',
    'author_email': 'woto@kaizoku.cyou',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ALiwoto/wotoplatform-Py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



#  WotoPlatform-Py - A Python library for interacting with WotoPlatform API.
#  Copyright (C) 2021-2022  ALiwoto - <woto@kaizoku.cyou> <https://github.com/ALiwoto>
#
#  This file is part of WotoPlatform-Py.
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


from .general import (
    ErrorCode,
    EndpointError,
    ServerException,
)

from .server_errors import *

__ERRORS_MAP__ = {
    ErrorCode.NoError: None,
    ErrorCode.ErrUnknown: UnknownError,
    ErrorCode.ErrServerUnavailable: ServerUnavailable,
    ErrorCode.ErrInvalidUsernameFormat: InvalidUsernameFormat,
    ErrorCode.ErrInvalidPasswordFormat: InvalidPasswordFormat,
    ErrorCode.ErrUsernameExists: UsernameExists,
    ErrorCode.ErrWrongUsername: WrongUsername,
    ErrorCode.ErrWrongPassword: WrongPassword,
    ErrorCode.ErrInvalidAuthKeyFormat: InvalidAuthKeyFormat,
    ErrorCode.ErrInvalidAccessHashFormat: InvalidAccessHashFormat,
    ErrorCode.ErrWrongAuthKey: WrongAuthKey,
    ErrorCode.ErrLoginAccessHashExpired: LoginAccessHashExpired,
    ErrorCode.ErrInvalidFirstName: InvalidFirstName,
    ErrorCode.ErrInvalidLastName: InvalidLastName,
    ErrorCode.ErrInvalidTitle: InvalidTitle,
    ErrorCode.ErrAlreadyAuthorized: AlreadyAuthorized,
    ErrorCode.ErrNotAuthorized: NotAuthorized,
    ErrorCode.ErrNotModified: NotModified,
    ErrorCode.ErrBioTooLong: BioTooLong,
    ErrorCode.ErrFirstNameTooLong: FirstNameTooLong,
    ErrorCode.ErrLastNameTooLong: LastNameTooLong,
    ErrorCode.ErrInvalidUsernameAndUserId: InvalidUsernameAndUserId,
    ErrorCode.ErrMethodNotImplemented: MethodNotImplemented,
    ErrorCode.ErrPermissionDenied: PermissionDenied,
    ErrorCode.ErrKeyNotFound: KeyNotFound,
    ErrorCode.ErrInvalidTelegramId: InvalidTelegramId,
    ErrorCode.ErrInvalidEmail: InvalidEmail,
}

def parse_server_error(error: EndpointError) -> ServerException:
    the_type = __ERRORS_MAP__.get(ErrorCode(error.code), None)
    if not isinstance(the_type, type):
        return UnknownError(error.message, error.origin)
    
    return the_type(error.message, error.origin)




#  WotoPlatform-Py - A Python library for interacting with WotoPlatform API.
#  Copyright (C) 2021-2022  ALiwoto - <woto@kaizoku.cyou> <https://github.com/ALiwoto>
#
#  This file is part of WotoPlatform-Py.
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
from .types.errors import (
    EmptyServerResponse,
    DataReceiveTimeout,
)


def make_sure_byte(b: bytes, length: int) -> bytes:
    """
    make_sure_byte is a useful function. consider b is a number in string
    and you are sending it as a byte to this function.
    this function then, will ensure that the length of this byte array
    is exactly equal to the passed-by argument.
    for example:
    make_sure_byte([]byte("5"), 8) will return []byte("5       ")
    the returned value's length will be exactly the same as length.
    """
    tmp_b = b
    while len(b) < length: tmp_b = tmp_b + b' '
    return bytes(tmp_b, 'utf-8')

class DataReceiver:
    received_data = None
    base_client = None
    _data_waiter: asyncio.Lock = None
    
    def __init__(self, base_client) -> None:
        self._data_waiter = asyncio.Lock()
        self.base_client = base_client
    
    async def wait_for_data(self, timeout: float = 20) -> bytes:
        try:
            await asyncio.wait_for(self._wait_for_data(), timeout=timeout)
        except asyncio.TimeoutError as e:
            raise DataReceiveTimeout(e)
        
        # await self._wait_for_data()
        
        if self.received_data == None:
            raise EmptyServerResponse()
        
        return self.received_data
    
    async def first_wait(self) -> None:
        if self._data_waiter.locked(): return
        await self._data_waiter.acquire()
        
    async def _wait_for_data(self) -> None:
        # while not self.received_data and not getattr(self.base_client, 'is_closed', False):
            # await asyncio.sleep(0.1)
        await self._data_waiter.acquire()
    
    def receive_data(self, data) -> None:
        self.received_data = data
        self._data_waiter.release()
    
    
    


#  WotoPlatform-Py - A Python library for interacting with WotoPlatform API.
#  Copyright (C) 2021-2022  ALiwoto - <woto@kaizoku.cyou> <https://github.com/ALiwoto>
#
#  This file is part of WotoPlatform-Py.
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
import typing
from pydantic import BaseModel

from wotoplatform.types.errors.general import EndpointError
from ..errors import ServerException
from ..errors import parse_server_error

__BATCH_STR__ = 'batch?='

class Scaffold():
    def get_action(self) -> int:
        pass
    
    def get_single_batch(self) -> str:
        pass

    def get_batch_execution(self) -> str:
        return __BATCH_STR__ + self.get_single_batch()
    
    def is_empty_scaffold(self):
        return False


class ScaffoldHolder():
    """
    ScaffoldHolder is a simple scaffold container which has unique-id.
    """
    _data_unique_id: str = ''
    scaffold_data: Scaffold = None
    def get_unique_id(self) -> str:
        return self._data_unique_id
    
    def set_unique_id(self, uid: str):
        self._data_unique_id = uid
    
    def __init__(self, unique_id: str, the_data: Scaffold) -> None:
        self._data_unique_id = unique_id
        self.scaffold_data = the_data
    
    def get_as_bytes(self) -> bytes:
        final_value = {
            'unique_id': self.get_unique_id(),
            'action': self.scaffold_data.get_action(),
            'batch_execute': self.scaffold_data.get_batch_execution(),
            'data': self.scaffold_data.json(),
        }
        return json.dumps(final_value).encode('utf-8')

class DScaffold(Scaffold, BaseModel):
    """
    DScaffold is a base class for all data classes that need to be sent to the server.
    """
    pass
    
    def get_response_type(self) -> type:
        pass
    


class EmptyScaffoldData(DScaffold):
    """
    EmptyScaffoldData describes an empty `DScaffold`.
    """
    def is_empty_scaffold(self):
        return True
    

class ResultScaffold(BaseModel):
    """
    ResultScaffold is a base class for all data classes that need to be received
        from the server as the result itself.
    """

    def __str__(self) -> str:
        try:
            return self.json(indent=4, ensure_ascii=False)
        except Exception: return super().__str__()

class RScaffold(Scaffold, BaseModel):
    """
    RScaffold is a base class for all data classes that need to be received 
        from the server as the response.
    """
    success: bool
    error: typing.Optional[EndpointError]
    result: typing.Optional[ResultScaffold]
    def get_action(self) -> int:
        pass

    def get_single_batch(self) -> str:
        pass

    def has_exception(self) -> bool:
        return self.error is not None
    
    def get_exception(self) -> ServerException:
        if self.has_exception():
            return parse_server_error(self.error)
        return None



class RawResponse(RScaffold):
    """
    A raw server response.
    """
    result: typing.Optional[typing.Union[dict, list]]


class RawDScaffold(Scaffold):
    """
    RawDScaffold is a Scaffold which contains raw data in itself, instead of having
    attributes like normal Scaffold classes.
    """
    __scaffold_action: int = 0
    __scaffold_single_batch: str = ''
    __scaffold_json_data: str = ''
    
    def __init__(self, action: int, the_batch: str, data: str) -> None:
        self.__scaffold_action = action
        self.__scaffold_single_batch = the_batch
        self.__scaffold_json_data = data
    
    def get_action(self) -> int:
        return self.__scaffold_action
    
    def get_single_batch(self) -> str:
        return self.__scaffold_single_batch
    
    def json(self) -> str:
        return self.__scaffold_json_data
    
    def parse_response(self, j_value):
        if j_value is dict:
            my_result = j_value['result']
            j_value['result'] = None
            raw_resp = RawResponse(**j_value)
            raw_resp.result = my_result
            return raw_resp
        return RawResponse(**j_value)

    def get_response_type(self) -> type:
        return RawResponse




#  WotoPlatform-Py - A Python library for interacting with WotoPlatform API.
#  Copyright (C) 2021-2022  ALiwoto - <woto@kaizoku.cyou> <https://github.com/ALiwoto>
#
#  This file is part of WotoPlatform-Py.
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
from .general import ClientException


class InvalidTypeException(ClientException):
    """
    Invalid type exception.
    """
    def __init__(self, expected_type: type, actual_type: type):
        self.expected_type = expected_type
        self.actual_type = actual_type
        self.message = f'Expected type of {expected_type}, got {actual_type}.'

class ClientNotInitializedException(ClientException):
    """
    Client not initialized exception.
    """
    def __init__(self):
        self.message = 'Client is not initialized.'

class ClientAlreadyInitializedException(ClientException):
    """
    Client already initialized exception.
    """
    def __init__(self):
        self.message = 'Client is already initialized.'

class ClientVersionNotAcceptable(ClientException):
    """
    Client version is not acceptable.
    """
    def __init__(self):
        self.message = f'Client version is not acceptable.'

class DataReceiveTimeout(ClientException):
    inner_exception: asyncio.TimeoutError = None
    """
    Client version is not acceptable.
    """
    def __init__(self, inner_exception: asyncio.TimeoutError = None):
        self.message = f'Receiving data from server timed out.'
        self.inner_exception = inner_exception

class EmptyServerResponse(ClientException):
    """
    Client version is not acceptable.
    """
    def __init__(self):
        self.message = f'Received empty data from server, most likely connection is closed.'

